package net.sf.sido.pojo;

import com.google.common.base.Function;
import net.sf.sido.DataFactory;
import net.sf.sido.DataObject;
import net.sf.sido.DataSchema;
import net.sf.sido.io.json.JSONDataObjectReader;
import org.apache.commons.lang.Validate;
import org.apache.commons.lang.reflect.ConstructorUtils;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.HashMap;
import java.util.Map;

public class DOSchema {

    protected final DOFactory doFactory;
    protected final DataSchema dataSchema;

    private final Map<String,Class<? extends DOObject>> classes = new HashMap<String,Class<? extends DOObject>>();

    private final JSONDataObjectReader reader = new JSONDataObjectReader();

    public DOSchema(DOFactory doFactory, String dataSchemaName) {
        this.doFactory = doFactory;
        this.dataSchema = doFactory.getDataFactory().getSchema(dataSchemaName);
        // Registration
        this.doFactory.register(this);
    }

    public DataFactory getDataFactory() {
        return doFactory.getDataFactory();
    }

    public DOFactory getDoFactory() {
        return doFactory;
    }

    public DataSchema getDataSchema() {
        return dataSchema;
    }

    public DataObject newInstance (String typeName) {
        return getDataFactory().newInstance(typeName);
    }

    @Override
    public String toString() {
        return dataSchema.getName();
    }

    protected void registerClass(String typeName, Class<? extends DOObject> cls) {
        classes.put(typeName, cls);
    }

    public <D extends DOObject> Function<DataObject, D> createFn() {
        return new Function<DataObject, D>() {
            @Override
            public D apply(DataObject o) {
                return (D) create(o);
            }
        };
    }

    public <D extends DOObject> D create (DataObject o) {
        if (o == null) {
            return null;
        } else {
            String type = o.getSidoType().toString();
            Class<D> cls = (Class<D>) classes.get(type);
            Validate.notNull(cls, "Cannot create POJO from type " + type);
            try {
                return (D) ConstructorUtils.invokeConstructor(cls, new Object[] {this, o});
            } catch (Exception e) {
                throw new IllegalStateException("Cannot create POJO", e);
            }
        }
    }

    public DataObject read(String type, String json) {
        try {
        InputStream in = new ByteArrayInputStream(json.getBytes("UTF-8"));
        try {
            DataFactory dataFactory = getDataFactory();
            return reader.read(in, dataFactory, dataFactory.getMandatoryType(type));
        } finally {
            in.close();
        }} catch (IOException ex) {
            throw new IllegalStateException(String.format("Cannot read %s type from JSON", type), ex);
        }
    }

    public <S extends DOSchema> S getSchema (Class<S> schemaClass) {
        return doFactory.getSchema(schemaClass);
    }
}
